# Copyright (c) 2021 Leonardo Uieda.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""Misc utilities."""
import contextlib
import datetime
import subprocess

from ._version import __version__ as nene_version


def generate_identifier(path):
    """
    Generate a unique identifier for the given path.

    Parameters
    ----------
    path : :class:`pathlib.Path`
        The path of the source file.

    Returns
    -------
    identifier : str
        String of ``/`` separated parents and stem of the path.

    """
    identifier = "/".join([str(i) for i in path.parts[:-1]] + [path.stem])
    return identifier


def capture_build_info():
    """
    Create a dictionary with information about the build environment.

    Returns
    -------
    build : dict
        Dictionary with the captured information.
    """
    build = {
        "today": datetime.datetime.utcnow(),
        "nene_version": nene_version,
    }
    # Have to be careful if git isn't installed or if not running in a repository
    with contextlib.suppress(Exception):
        git_output = subprocess.run(
            ["git", "rev-parse", "--short", "HEAD"], capture_output=True, text=True
        )
        commit = git_output.stdout.strip()
        if commit:
            build["commit"] = commit
    return build



# Copyright (c) 2021 Leonardo Uieda.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""The Click command line interface ("main" function)."""
import sys

import click

from . import _api
from .printing import make_console

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}
DEFAULT_CONFIG = "config.yml"


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config",
    "-c",
    default=DEFAULT_CONFIG,
    show_default=True,
    help="Specify a different configuration file",
)
@click.option(
    "--serve",
    "-s",
    default=False,
    is_flag=True,
    show_default=True,
    help="Serve the built website and watch for changes (opens your browser)",
)
@click.option(
    "--verbose/--quiet",
    "-v/-q",
    default=True,
    show_default=True,
    help="Print information during execution / Don't print",
)
@click.version_option()
def main(config, serve, verbose):
    """
    Nene: A no-frills static site generator.

    Builds a static HTML website from sources, templates, and configuration
    found in the current directory.
    """
    config_file = config
    console, style = make_console(verbose)
    style_rule = "green"
    console.rule(
        f":palm_tree: [{style_rule}]Lay back and relax while Nēnē builds "
        "your website[/] :palm_tree:",
        style=style,
    )
    # Main website building section
    try:
        with console.status(f"[{style}]Working...[/{style}]"):
            site, source_files, config, build = _api.build(config_file, console, style)
            _api.render(site, config, build, console, style)
            _api.export(
                site, source_files["copy"], config["output_dir"], console, style
            )
    # Handle exceptions nicely when they occur
    except Exception:
        console_error, style_error = make_console(verbose=True, style="bold red")
        console.print()
        console.rule(":fire: Error messages start here :fire:", style=style_error)
        console.print_exception(suppress=[click])
        console.rule(":fire: Error messages above :fire:", style=style_error)
        console.print()
        console_error.print(
            ":pensive: Oh no! Something went wrong while building your website.",
            style=style_error,
        )
        if not verbose:
            console_error.print(
                ":bulb: "
                "You may want to run Nēnē again with verbosity turned on ('--verbose') "
                "to get more information about what happened."
            )
        console.print()
        console.print("A few things to check:", style="bold")
        console.print("  1. The error messages above for clues.")
        console.print("  2. Configuration and data files and Markdown headers.")
        console.print()
        console.print("Can't figure out what is happening?", style="bold")
        console.print(
            "  * Consider submitting a bug report: "
            "https://github.com/leouieda/nene/issues"
        )
        console.print("  * Please include the output above :point_up_2: when doing so.")
        console.print()
        sys.exit(1)
    else:
        if not serve:
            console.rule(f":tada: [{style_rule}]Done![/] :tada:", style=style)
    # Start the livereload server is requested
    if serve:
        console.rule(
            f":robot: [{style_rule}]Starting development server[/] :robot:",
            style=style,
        )
        _api.serve(
            config,
            source_files,
            console=console,
            style=style,
        )
    sys.exit(0)



# Copyright (c) 2021 Leonardo Uieda.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""Version string coming from setuptools_scm."""
from . import _version_generated

# Create this in a separate module to avoid circular imports when placed in
# nene/__init__.py
__version__ = f"v{_version_generated.version}"



# Copyright (c) 2021 Leonardo Uieda.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""
Nēnē: A no-frills static site generator.
"""
from ._api import build, export, render, serve
from ._version import __version__



# Copyright (c) 2021 Leonardo Uieda.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""Utilities for printing things with the Rich Console."""

import rich.console


def make_console(verbose, style="bold blue"):
    """
    Start up the :class:`rich.console.Console` instance we'll use.

    Parameters
    ----------
    verbose : bool
        Whether or not to print status messages to stderr.
    """
    console = rich.console.Console(stderr=True, quiet=not verbose, highlight=False)
    return console, style


def print_dict(dictionary, console):
    """
    Print the key/value pairs of a dictionary one per line.

    Parameters
    ----------
    dictionary : dict
        The dict.
    console : rich.console.Console
        The console used for printing.
    """
    for key in sorted(dictionary):
        value = dictionary[key]
        if isinstance(value, str):
            value = " ".join([part.strip() for part in value.split("\n")])
        console.print(f"   {key}: {value}", highlight=False)


def print_file_stats(files, console):
    """
    Print the number of files of each type.

    Parameters
    ----------
    files : dict
        Dictionary with lists of files separated by file type.
    console : rich.console.Console
        The console used for printing.
    """
    line_length = max(len(key) for key in files)
    for file_type in files:
        file_type_print = " " * (line_length - len(file_type)) + file_type
        console.print(f"   {file_type_print} = {len(files[file_type])}")



# Copyright (c) 2021 Leonardo Uieda.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""Render outputs with Jinja templates."""
from pathlib import Path

import jinja2
import mdit_py_plugins.anchors
import mdit_py_plugins.footnote
from markdown_it import MarkdownIt


def make_jinja_envs(templates_dir):
    """
    Create the default Jinja environments given the template folder.

    Parameters
    ----------
    templates_dir : str or pathlib.Path
        The path to the templates.

    Returns
    -------
    envs : dict
        The keys are the file types supported as templates and values are the
        corresponding environments used to render the templates for those file
        types.
    """
    envs = {
        "markdown": jinja2.Environment(
            loader=jinja2.FileSystemLoader([str(templates_dir)], followlinks=True),
            extensions=["jinja2.ext.do", "jinja2.ext.loopcontrols"],
        ),
        "html": jinja2.Environment(
            loader=jinja2.FileSystemLoader([str(templates_dir)], followlinks=True),
            extensions=["jinja2.ext.do", "jinja2.ext.loopcontrols"],
        ),
        "latex": jinja2.Environment(
            loader=jinja2.FileSystemLoader([str(templates_dir)], followlinks=True),
            extensions=["jinja2.ext.do", "jinja2.ext.loopcontrols"],
            # Define custom syntax that is compatible with LaTeX
            # Based on jtex by curvenote (MIT license):
            # https://github.com/curvenote/jtex/blob/2778c9fc51cd2cbbe8d4b7deedd637e9dd59f662/jtex/TemplateRenderer.py#L36
            block_start_string=r"[#",
            block_end_string="#]",
            variable_start_string=r"[-",
            variable_end_string="-]",
            line_comment_prefix=r"%%",
            comment_start_string=r"%#",
            comment_end_string="#%",
            trim_blocks=True,
            autoescape=False,
            auto_reload=True,
            keep_trailing_newline=True,
        ),
    }
    return envs


def markdown_to_html(page):
    """
    Convert the Markdown content of a page to HTML.

    Parameters
    ----------
    page : dict
        Dictionary with the parsed YAML front-matter and Markdown body.

    Returns
    -------
    html : str
        The converted HTML.

    """
    parser = MarkdownIt("commonmark", {"typographer": True})
    parser.enable(["replacements", "smartquotes"])
    parser.enable("table")
    parser.use(mdit_py_plugins.anchors.anchors_plugin)
    parser.use(mdit_py_plugins.footnote.footnote_plugin)
    # Remove the starting hr from the footnote block. It's ugly and should be
    # handled through CSS by putting a top border on section element.
    parser.add_render_rule("footnote_block_open", _render_footnote_block_open)
    html = parser.render(page["markdown"])
    return html


def _render_footnote_block_open(self, tokens, idx, options, env):
    """Render the footnote opening without the hr tag at the start."""
    html = mdit_py_plugins.footnote.index.render_footnote_block_open(
        self, tokens, idx, options, env
    )
    lines = html.split("\n")
    if lines[0].strip().startswith("<hr"):
        lines = lines[1:]
    return "\n".join(lines)


def render_markdown(page, config, site, build, jinja_envs):
    """
    Render the templates in Markdown content of the page.

    Parameters
    ----------
    page : dict
        Dictionary with the parsed YAML front-matter and Markdown body.
    config : dict
        A dictionary with the default configuration and variables loaded from
        the file.
    site : dict
        Dictionary with the entire site content so far.
    build : dict
        Dictionary with information about the build environment.
    jinja_envs
        A dictionary of Jinja2 environments for loading templates.

    Returns
    -------
    markdown : str
        The rendered Markdown content for the page.
    """
    template = jinja_envs["markdown"].from_string(page["markdown"])
    markdown = template.render(page=page, config=config, site=site, build=build)
    return markdown


def render_output(page, config, site, build, jinja_envs):
    """
    Render the full template output for a page.

    Parameters
    ----------
    page : dict
        Dictionary with the page metadata and body content (in HTML not just
        Markdown).
    config : dict
        A dictionary with the default configuration and variables loaded from
        the file.
    site : dict
        Dictionary with the entire site content so far.
    build : dict
        Dictionary with information about the build environment.
    jinja_envs
        A dictionary of Jinja2 environments for loading templates.

    Returns
    -------
    html : str
        The converted HTML.

    """
    types = {".html": "html", ".tex": "latex"}
    template_type = types[Path(page["template"]).suffix]
    template = jinja_envs[template_type].get_template(page["template"])
    html = template.render(page=page, config=config, site=site, build=build)
    return html



# Copyright (c) 2021 Leonardo Uieda.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""Find files by crawling a directory tree."""
from glob import glob
from pathlib import Path


def _walk_non_hidden(root, hidden_start=(".", "_")):
    """
    Walk a directory tree while ignoring paths that start with some markers.

    Parameters
    ----------
    root : str or :class:`pathlib.Path`
        The base path to start crawling.
    hidden_start : tuple of str
        List of starting characters to ignore.

    Yields
    ------
    path : str or :class:`pathlib.Path`
        Path to a file.
    """
    for path in Path(root).glob("*"):
        if not any(path.name.startswith(hidden) for hidden in hidden_start):
            if path.is_dir():
                yield from _walk_non_hidden(path, hidden_start)
            else:
                yield path


def crawl(root, ignore, copy_extra):
    """
    Crawl the directory root separating inputs by type and files for copying.

    Paths starting with `_` and `.` or in *ignore* will not be included.

    Parameters
    ----------
    root : str or :class:`pathlib.Path`
        The base path to start crawling.
    ignore : list of str
        List of paths to ignore when crawling. Paths can contain globbing syntax.
    copy_extra : list of str
        List of extra paths to copy to the build directory.

    Returns
    -------
    tree : dict
        A dictionary with keys `"markdown"`, `"ipynb"`, `"json"`, and `"copy"`
        containing lists of Paths that fall in each category.

    """
    formats = {
        ".md": "markdown",
        ".json": "json",
        ".yml": "yaml",
        ".yaml": "yaml",
        ".ipynb": "ipynb",
        ".bib": "bibtex",
    }
    tree = {
        "copy": [Path(path) for path in copy_extra],
        "markdown": [],
        "ipynb": [],
        "json": [],
        "yaml": [],
        "bibtex": [],
    }
    expanded_ignore = []
    for path in ignore:
        expanded_ignore.extend(glob(path, recursive=True))
    for path in _walk_non_hidden(root):
        if str(path) in expanded_ignore:
            continue
        if path.suffix in formats:
            tree[formats[path.suffix]].append(path)
        else:
            tree["copy"].append(path)
    return tree



# Copyright (c) 2021 Leonardo Uieda.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""Public functions used for building and serving the website."""
import itertools
import logging
import shutil
from pathlib import Path

import livereload

from .crawling import crawl
from .parsing import load_config, load_data, load_jupyter_notebook, load_markdown
from .printing import make_console, print_dict, print_file_stats
from .rendering import make_jinja_envs, markdown_to_html, render_markdown, render_output
from .utils import capture_build_info


def build(config_file, console=None, style=""):
    """
    Build the website structure from the sources.

    Reads in configuration, page sources, and data sources. Structures the
    content into dictionaries and prepares the website for rendering.

    Parameters
    ----------
    config_file : str
        Path to the configuration file.
    console : rich.console.Console
        Console used to print status messaged. If None, no messages are
        printed.
    style : str
        Style string used to format console status messages.

    Returns
    -------
    site : dict
        The generated website as a dictionary of pages. Each page has a unique
        ID (usually the relative file path without extension) and is a
        dictionary.
    source_files : dict
        Dictionary with lists of source files by file type.
    config : dict
        Parameters read from the main configuration file.
    build: dict
        Information about the current build (version of Nēnē, git commit, date,
        etc).
    """
    if console is None:
        console, style = make_console(verbose=False)

    console.print(
        ":package: Captured information about the build environment:", style=style
    )
    build_info = capture_build_info()
    print_dict(build_info, console)

    console.print(f":package: Configuration loaded from '{config_file}':", style=style)
    config = load_config(config_file)
    print_dict(config, console)

    console.print(":open_file_folder: Scanned source directory:", style=style)
    source_files = crawl(
        root=Path("."), ignore=config["ignore"], copy_extra=config["copy"]
    )
    print_file_stats(source_files, console)

    site = {}

    console.print(":open_book: Reading Markdown files:", style=style)
    if source_files["markdown"]:
        for path in source_files["markdown"]:
            console.print(f"   {str(path)}")
            page = load_markdown(path)
            site[page["id"]] = page
    else:
        console.print("   None found.")

    console.print(
        ":open_book: Reading Jupyter Notebook files as Markdown:", style=style
    )
    if source_files["ipynb"]:
        for path in source_files["ipynb"]:
            console.print(f"   {str(path)}")
            page = load_jupyter_notebook(path)
            site[page["id"]] = page
    else:
        console.print("   None found.")

    console.print(
        ":open_book: Reading data files and propagating to pages:", style=style
    )
    data_files = source_files["json"] + source_files["yaml"] + source_files["bibtex"]
    if data_files:
        for path in sorted(data_files):
            console.print(f"   {str(path)}")
            data = load_data(path)
            if data["id"] in site and not data["id"].endswith("/index"):
                propagate_to_page = [site[data["id"]]]
            else:
                propagate_to_page = [
                    page for page in site.values() if page["parent"] == data["parent"]
                ]
            for page in propagate_to_page:
                console.print(f"     ↳ {page['id']}")
                if data["id"].endswith("/index") or page["id"] == data["id"]:
                    # Merge the two data dictionaries with 'page' taking precedence
                    page.update({**data["content"], **page})
                else:
                    page[data["id"].split("/")[-1]] = data["content"]
    else:
        console.print("   None found.")

    console.print(":dart: Determining destination path for pages:", style=style)
    for page in site.values():
        if "save_as" in page:
            destination = Path(page["source"]).with_name(page["save_as"])
        else:
            destination = Path(page["source"]).with_suffix(".html")
        page["path"] = str(destination)
        console.print(f"   {page['id']} ⇒ {page['path']}")

    def get_parent(item):
        """Return the name of the parent of a page for the groupby."""
        page = item[1]
        return page["parent"]

    console.print(":baby: Adding information on sibling pages:", style=style)
    grouped = itertools.groupby(sorted(site.items(), key=get_parent), key=get_parent)
    for parent, group in grouped:
        if parent == ".":
            console.print("   . (base)")
        else:
            console.print(f"   {parent}")
        # The groups are also (key, value) pairs like dict.items().
        siblings = {key for key, value in group}
        for page_id in siblings:
            console.print(f"     ↳ {page_id}")
            site[page_id]["siblings"] = [site[i] for i in (siblings - {page_id})]

    return site, source_files, config, build_info


def render(site, config, build, console=None, style=""):
    """
    Render the HTML or other outputs from the assembled website sources.

    The ``site``, ``config``, and ``build`` variables are passed to the Jinja2
    template for rendering the final outputs.

    Modifies the pages in ``site`` **in place** to add the rendered output of
    each page.

    Parameters
    ----------
    site : dict
        The generated website as a dictionary of pages. Each page has a unique
        ID (usually the relative file path without extension) and is a
        dictionary. The rendered output of each page is added to
        ``page["output"]`` **in place**..
    config : dict
        Parameters read from the main configuration file.
    build: dict
        Information about the current build (version of Nēnē, git commit, date,
        etc).
    console : rich.console.Console
        Console used to print status messaged. If None, no messages are
        printed.
    style : str
        Style string used to format console status messages.
    """
    if console is None:
        console, style = make_console(verbose=False)

    jinja_envs = make_jinja_envs(config["templates_dir"])

    console.print(":art: Rendering templates in Markdown content:", style=style)
    for page in site.values():
        console.print(f"   {page['source']}")
        page["markdown"] = render_markdown(page, config, site, build, jinja_envs)

    console.print(":art: Converting Markdown content to HTML:", style=style)
    for page in site.values():
        if "markdown" in page:
            console.print(f"   {page['source']}")
            page["body"] = markdown_to_html(page)

    console.print(":art: Rendering templates for final outputs:", style=style)
    for page in site.values():
        console.print(f"   {page['path']} ← {page['template']}")
        page["output"] = render_output(page, config, site, build, jinja_envs)


def export(site, files_to_copy, output_dir, console=None, style=""):
    """
    Write the output (HTML) and other website files to disk.

    To be used after running ``build``.

    Parameters
    ----------
    site : dict
        The generated website as a dictionary of pages.
    files_to_copy : list
        List of paths to files that need to be copied to the output folder.
    output_dir : str or pathlib.Path
        Path of the output folder.
    console : rich.console.Console
        Console used to print status messaged. If None, no messages are
        printed.
    style : str
        Style string used to format console status messages.
    """
    if console is None:
        console, style = make_console(verbose=False)

    output_dir = Path(output_dir)
    output_dir.mkdir(exist_ok=True)

    console.print(
        ":deciduous_tree: Copying source directory tree and extra files to "
        f"'{str(output_dir)}':",
        style=style,
    )
    for path in files_to_copy:
        console.print(f"   {str(path)}")
        destination = output_dir / path
        destination.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(path, destination)

    console.print(
        f":writing_hand:  Writing output files to '{str(output_dir)}':", style=style
    )
    for page in site.values():
        destination = output_dir / Path(page["path"])
        console.print(f"   {str(destination)} ⇒  id: {page['id']}")
        destination.parent.mkdir(parents=True, exist_ok=True)
        destination.write_text(page["output"], encoding="utf-8")

    console.print(":bar_chart: Writing Jupyter Notebook image files:", style=style)
    pages_with_images = [
        page
        for page in site.values()
        if "images" in page and page["images"] is not None
    ]
    if pages_with_images:
        for page in pages_with_images:
            console.print(f"   {page['source']}")
            for image_path, image in page["images"].items():
                path = output_dir / Path(page["parent"]) / Path(image_path)
                console.print(f"     ↳ {str(path)}")
                path.parent.mkdir(parents=True, exist_ok=True)
                path.write_bytes(image)
    else:
        console.print("   None found.")

    console.print(
        f":gift: Your built website is in '{str(output_dir)}'. Enjoy!",
        style=style,
    )


def serve(config, source_files, port=None, console=None, style=""):
    """
    Serve the output folder with livereload and watch the sources for changes.

    When any of the source files are changed, the website is rebuilt
    automatically.

    Parameters
    ----------
    config : dict
        Parameters read from the main configuration file.
    source_files : dict
        Dictionary with lists of source files by file type.
    port : int
        The port used by the server. If None, will first try port 5500 and
        increase by 1 if the port is already busy (maximum of 10 times).
    console : rich.console.Console
        Console used to print status messaged. If None, no messages are
        printed.
    style : str
        Style string used to format console status messages.
    """
    # Note: Can't really quite the livereload server since it sets the logging
    # level and handler when "serve" is called. So we can't set it before
    # anything happens. See livereload/server.py line 331 (as of 2022-04-08).
    livereload_logger = logging.getLogger("livereload")
    if console is None:
        console, style = make_console(verbose=False)

    config_file = config["config_file"]
    watch = [config_file, config["templates_dir"]]
    for category in source_files:
        watch.extend(source_files[category])
    # Make all paths into strings to avoid "JSON serialization errors" coming
    # from livereload
    watch = sorted([str(fname) for fname in watch])

    def rebuild():
        """Rebuild the website."""
        site, source_files, config, build_info = build(config_file)
        render(site, config, build_info)
        export(site, source_files["copy"], config["output_dir"])

    console.print(":eyes: Watching these source files for changes:", style=style)
    server = livereload.Server()
    for fname in watch:
        console.print(f"   {fname}")
        server.watch(fname, rebuild)

    if port is None:
        max_attempts = 10
        port = 5500
    else:
        max_attempts = 1
    for attempt in range(max_attempts):
        console.print(":ledger: Server log messages:", style=style)
        try:
            server.serve(
                root=str(config["output_dir"]),
                host="localhost",
                port=port,
                open_url_delay=1,
            )
            break
        except OSError:
            if attempt == max_attempts + 1:
                raise
            console.print(
                "   Failed to start server because "
                f"port {port} is already in use. "
                "Trying another port."
            )
            port += 1
            # Clear the logger to avoid having livereload duplicate the
            # handlers causing multiple prints of the same log messages.
            livereload_logger.handlers.clear()



# Copyright (c) 2021 Leonardo Uieda.
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT
"""Load data from source files."""
import json
from pathlib import Path

import yaml

# For Jupyter notebook support
try:
    import nbconvert
    import nbformat
    import traitlets.config
except ImportError:
    traitlets = None
    nbconvert = None
    nbformat = None

# For bibtex support
try:
    import bibtexparser
    from bibtexparser.bparser import BibTexParser
    from bibtexparser.customization import convert_to_unicode

except ImportError:
    bibtexparser = None

from .utils import generate_identifier


def _read_data_file(path):
    """
    Read the contents of a data file in JSON, YAML or BibTeX format.

    Parameters
    ----------
    path : :class:`pathlib.Path`
        The path of the JSON or YAML data file.

    Returns
    -------
    data : dict
        The contents of the file as a dictionary.
    """
    path = Path(path)
    suffix = path.suffix
    loader = {".yml": yaml.safe_load, ".yaml": yaml.safe_load, ".json": json.loads}
    data = loader[suffix](path.read_text(encoding="utf-8"))
    return data


def load_bibtex(path):
    """
    Read bibtex entries from path.

    Parameters
    ----------
    path : :class:`pathlib.Path`
        The path of the bibtex file.

    Returns
    -------
    data : dict
        The publication entries of the bibfile as a dictionary.
    """
    parser = BibTexParser()
    parser.customization = convert_to_unicode
    with open(path) as bibtex_file:
        bibtex_database = bibtexparser.load(bibtex_file, parser=parser)
    return bibtex_database.entries


def load_config(fname):
    """
    Load configuration from a JSON or YAML file and append to the defaults.

    Parameters
    ----------
    fname : str or :class:`pathlib.Path`
        The name or path of the configuration file (has to be in JSON or YAML
        format).

    Returns
    -------
    config : dict
        A dictionary with the default configuration and variables loaded from
        the file.
    """
    config = {
        "ignore": [],
        "output_dir": "_build",
        "templates_dir": "_templates",
        "copy": [],
        "config_file": str(fname),
    }
    config.update(_read_data_file(fname))
    config["ignore"].append(fname)
    return config


def load_data(path):
    """
    Read JSON and YAML data from path.

    Parameters
    ----------
    path : :class:`pathlib.Path`
        The path of the JSON or YAML data file.

    Returns
    -------
    data : dict
        Dictionary with the file ID in ``"id"`` and the data in ``"content"``.
    """
    identifier = generate_identifier(path)
    if bibtexparser is not None and path.suffix == ".bib":
        content = {"bibtex_" + path.stem: load_bibtex(path)}
    else:
        content = _read_data_file(path)
    data = {
        "id": identifier,
        "type": "data",
        "parent": str(path.parent),
        "source": str(path),
        "content": content,
    }
    return data


def load_markdown(path):
    """
    Read Markdown content from path, including YAML front matter.

    Parameters
    ----------
    path : :class:`pathlib.Path`
        The path of the Markdown file.

    Returns
    -------
    page : dict
        Dictionary with the parsed YAML front-matter and Markdown body.
    """
    identifier = generate_identifier(path)
    page = {
        "id": identifier,
        "type": "markdown",
        "parent": str(path.parent),
        "source": str(path),
    }
    text = path.read_text(encoding="utf-8").strip()
    _, front_matter, markdown = text.split("---", maxsplit=2)
    page.update(yaml.safe_load(front_matter.strip()))
    page["markdown"] = markdown
    return page


def load_jupyter_notebook(path):
    """
    Read Jupyter notebook content as Markdown.

    Parameters
    ----------
    path : :class:`pathlib.Path`
        The path of the notebook file.

    Returns
    -------
    page : dict
        Dictionary with notebook content converted to HTML.
    """
    if nbconvert is None or nbformat is None:
        raise ValueError("Need nbconvert and nbformat to read Jupyter notebook files.")
    identifier = generate_identifier(path)
    page = {
        "id": identifier,
        "type": "ipynb",
        "parent": str(path.parent),
        "source": str(path),
    }
    notebook = nbformat.reads(path.read_text(encoding="utf-8"), as_version=4)
    image_dir = str(path.stem) + "_images"
    nbconfig = traitlets.config.Config()
    nbconfig.ExtractOutputPreprocessor.output_filename_template = str(
        Path(image_dir) / "{unique_key}_{cell_index}_{index}{extension}"
    )
    exporter = nbconvert.MarkdownExporter(config=nbconfig)
    markdown, resources = exporter.from_notebook_node(notebook)
    page["markdown"] = markdown
    page["images"] = resources["outputs"]
    return page
